import asyncio
import logging
import os
import shutil
import signal
import sys
from traceback import format_exc

import numpy as np
from PySide6.QtCore import (
    QCommandLineOption,
    QCommandLineParser,
    QObject,
    QSettings,
    QStandardPaths,
    Qt,
    QTimer,
    Signal,
)
from PySide6.QtGui import QAction, QIcon, QPixmap
from PySide6.QtWidgets import (
    QApplication,
    QLabel,
    QMainWindow,
    QMenu,
    QSplashScreen,
    QStackedWidget,
    QSystemTrayIcon,
)
from PySide6_GlobalHotkeys import Listener, bindHotkeys
from voiceconversion.common.deviceManager.DeviceManager import DeviceManager
from voiceconversion.downloader.WeightDownloader import (
    CONTENT_VEC_500_ONNX,
    downloadWeight,
)
from voiceconversion.ModelSlotManager import ModelSlotManager
from voiceconversion.RVC.RVCModelSlotGenerator import (
    RVCModelSlotGenerator,  # Parameters cannot be obtained when imported at startup.
)
from voiceconversion.RVC.RVCr2 import RVCr2
from voiceconversion.utils.LoadModelParams import LoadModelParams
from voiceconversion.utils.VoiceChangerModel import AudioInOutFloat
from voiceconversion.VoiceChangerSettings import VoiceChangerSettings
from voiceconversion.VoiceChangerV2 import VoiceChangerV2

from .audio import Audio
from .customizeui import CustomizeUiWidget
from .exceptions import (
    FailedToSetModelDirException,
    PipelineNotInitializedException,
    VoiceChangerIsNotSelectedException,
)
from .windowarea import WindowAreaWidget

PRETRAIN_DIR_NAME = "pretrain"
MODEL_DIR_NAME = "model_dir"

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)-15s %(levelname)-8s [%(module)s] %(message)s",
    handlers=[stream_handler],
)

logger = logging.getLogger(__name__)

# The IDs to talk with the keybindings configurator about the voice cards.
VOICE_CARD_KEYBIND_ID_PREFIX = "voice_card_"

ENABLE_PASS_THROUGH_KEYBIND_ID = "enable_pass_through"
DISABLE_PASS_THROUGH_KEYBIND_ID = "disable_pass_through"


class MainWindow(QMainWindow):
    def initialize(self, modelDir: str):
        centralWidget = QStackedWidget()
        self.setCentralWidget(centralWidget)

        self.windowAreaWidget = WindowAreaWidget(modelDir)
        centralWidget.addWidget(self.windowAreaWidget)

        self.customizeUiWidget = CustomizeUiWidget()

        preferencesMenu = self.menuBar().addMenu("Preferences")

        custumizeUiAction = QAction("Customize...", self)
        custumizeUiAction.triggered.connect(
            lambda: centralWidget.setCurrentWidget(self.customizeUiWidget)
        )
        self.customizeUiWidget.back.connect(
            lambda: centralWidget.setCurrentWidget(self.windowAreaWidget)
        )
        centralWidget.addWidget(self.customizeUiWidget)

        centralWidget.setCurrentWidget(self.windowAreaWidget)

        preferencesMenu.addAction(custumizeUiAction)

        def onVoiceCardHotkey(shortcutId: str):
            if shortcutId.startswith(VOICE_CARD_KEYBIND_ID_PREFIX):
                rowPlusOne = shortcutId.removeprefix(VOICE_CARD_KEYBIND_ID_PREFIX)
                if rowPlusOne.isdigit():
                    row = int(rowPlusOne) - 1  # 1-based indexing
                    if (
                        # 1 placeholder card
                        row < self.windowAreaWidget.voiceCards.count() - 1
                        and row >= 0
                    ):
                        self.windowAreaWidget.voiceCards.setCurrentRow(row)
            elif shortcutId == ENABLE_PASS_THROUGH_KEYBIND_ID:
                self.windowAreaWidget.passThroughButton.setChecked(True)
            elif shortcutId == DISABLE_PASS_THROUGH_KEYBIND_ID:
                self.windowAreaWidget.passThroughButton.setChecked(False)

        self.hotkeyListener = Listener()
        self.hotkeyListener.hotkeyPressed.connect(onVoiceCardHotkey)

        configureKeybindingsAction = QAction("Configure Keybindings...", self)
        configureKeybindingsAction.triggered.connect(
            lambda: bindHotkeys(
                [
                    (
                        f"{VOICE_CARD_KEYBIND_ID_PREFIX}{row}",
                        {"description": f"Select Voice Card {row}"},
                    )
                    for row in range(
                        1,  # 1-based indexing
                        self.windowAreaWidget.voiceCards.count(),  # 1 placeholder card
                        1,
                    )
                ]
                + [
                    (
                        ENABLE_PASS_THROUGH_KEYBIND_ID,
                        {"description": "Enable Pass Through"},
                    ),
                    (
                        DISABLE_PASS_THROUGH_KEYBIND_ID,
                        {"description": "Disable Pass Through"},
                    ),
                ],
            )
        )

        preferencesMenu.addAction(configureKeybindingsAction)

        self.systemTrayIcon = QSystemTrayIcon(self.windowIcon(), self)
        systemTrayMenu = QMenu()
        activateWindowAction = QAction("Show AVoc", self)
        activateWindowAction.triggered.connect(
            lambda: self.windowHandle().requestActivate()
        )
        quitAction = QAction("Quit AVoc", self)
        quitAction.triggered.connect(lambda: QApplication.quit())
        systemTrayMenu.addActions([activateWindowAction, configureKeybindingsAction])
        systemTrayMenu.addSeparator()
        systemTrayMenu.addAction(quitAction)
        self.systemTrayIcon.setContextMenu(systemTrayMenu)
        self.systemTrayIcon.setToolTip(self.windowTitle())
        self.systemTrayIcon.show()

        self.vcm: VoiceChangerManager | None = (
            None  # TODO: remove the no-model-load CLI arg
        )

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.close()  # closes the window (quits the app if it's the last window)
        else:
            super().keyPressEvent(event)

    def showTrayMessage(self, title: str, msg: str, icon: QIcon | QPixmap):
        self.systemTrayIcon.showMessage(title, msg, icon, 1000)


class VoiceChangerManager(QObject):

    modelUpdated = Signal(int)
    modelSettingsLoaded = Signal(int, float, float)

    def __init__(self, modelDir: str, pretrainDir: str):
        super().__init__()

        self.modelDir = modelDir
        self.pretrainDir = pretrainDir
        self.audio: Audio | None = None

        voiceChangerSettings = self.getVoiceChangerSettings()
        self.passThrough = False

        self.modelSlotManager = ModelSlotManager.get_instance(
            self.modelDir, "upload_dir"
        )  # TODO: fix the dir

        self.device_manager = DeviceManager.get_instance()
        self.devices = self.device_manager.list_devices()
        self.device_manager.initialize(
            voiceChangerSettings.gpu,
            voiceChangerSettings.forceFp32,
            voiceChangerSettings.disableJit,
        )

        self.vc = VoiceChangerV2(voiceChangerSettings)

    def getVoiceChangerSettings(self):
        voiceChangerSettings = VoiceChangerSettings()
        audioSettings = QSettings()
        audioSettings.beginGroup("AudioSettings")
        interfaceSettings = QSettings()
        interfaceSettings.beginGroup("Interface")
        voiceChangerSettingsDict = {
            "version": "v1",
            "modelSlotIndex": int(interfaceSettings.value("currentVoiceCardIndex", -1)),
            "inputSampleRate": int(
                audioSettings.value("sampleRate")
            ),  # TODO: validation
            "outputSampleRate": int(
                audioSettings.value("sampleRate")
            ),  # TODO: validation
            "gpu": 0,
            "extraConvertSize": 0.1,
            "serverReadChunkSize": 22,
            "crossFadeOverlapSize": 0.1,
            "forceFp32": 0,
            "disableJit": 0,
            "enableServerAudio": 1,
            "exclusiveMode": 0,
            "asioInputChannel": -1,
            "asioOutputChannel": -1,
            "dstId": 0,
            "f0Detector": "rmvpe_onnx",
            "tran": 6,
            "formantShift": 0.0,
            "useONNX": 0,
            "silentThreshold": -90,
            "indexRatio": 0.0,
            "protect": 0.5,
            "silenceFront": 1,
        }
        voiceChangerSettings.set_properties(voiceChangerSettingsDict)
        return voiceChangerSettings

    def initialize(self):
        voiceChangerSettings = self.getVoiceChangerSettings()
        val = voiceChangerSettings.modelSlotIndex
        slotInfo = self.modelSlotManager.get_slot_info(val)
        if slotInfo is None or slotInfo.voiceChangerType is None:
            logger.warning(f"Model slot is not found {val}")
            return

        voiceChangerSettings.set_properties(
            {
                "tran": slotInfo.defaultTune,
                "formantShift": slotInfo.defaultFormantShift,
                "indexRatio": slotInfo.defaultIndexRatio,
                "protect": slotInfo.defaultProtect,
            }
        )

        self.modelSettingsLoaded.emit(
            slotInfo.defaultTune,
            slotInfo.defaultFormantShift,
            slotInfo.defaultIndexRatio,
        )

        if slotInfo.voiceChangerType == self.vc.get_type():
            self.vc.set_slot_info(
                slotInfo,
                self.pretrainDir,
            )
            # TODO: unify changing properties that don't need reinit.
            self.vc.vcmodel.settings.tran = slotInfo.defaultTune
            self.vc.vcmodel.settings.formantShift = slotInfo.defaultFormantShift
            self.vc.vcmodel.settings.indexRatio = slotInfo.defaultIndexRatio
        elif slotInfo.voiceChangerType == "RVC":
            logger.info("Loading RVC...")
            self.vc.initialize(
                RVCr2(
                    self.modelDir,
                    os.path.join(self.pretrainDir, CONTENT_VEC_500_ONNX),
                    slotInfo,
                    voiceChangerSettings,
                ),
                self.pretrainDir,
            )
        else:
            logger.error(f"Unknown voice changer model: {slotInfo.voiceChangerType}")

    def setModelSettings(
        self,
        pitch: int,
        formantShift: float,
        index: float,
    ):
        voiceChangerSettings = self.getVoiceChangerSettings()
        val = voiceChangerSettings.modelSlotIndex
        slotInfo = self.modelSlotManager.get_slot_info(val)
        if slotInfo is None or slotInfo.voiceChangerType is None:
            logger.warning(f"Model slot is not found {val}")
            return

        slotInfo.defaultTune = pitch
        slotInfo.defaultFormantShift = formantShift
        slotInfo.defaultIndexRatio = index

        if self.vc.vcmodel is not None:
            # TODO: unify changing properties that don't need reinit.
            self.vc.vcmodel.settings.tran = slotInfo.defaultTune
            self.vc.vcmodel.settings.formantShift = slotInfo.defaultFormantShift
            self.vc.vcmodel.settings.indexRatio = slotInfo.defaultIndexRatio

        self.modelSlotManager.save_model_slot(val, slotInfo)

    def setRunning(self, running: bool, passThrough: bool):
        if (self.audio is not None) == running:
            return

        if running:
            voiceChangerSettings = self.getVoiceChangerSettings()
            settings = QSettings()
            settings.beginGroup("AudioSettings")
            self.audio = Audio(
                settings.value("audioInputDevice"),
                settings.value("audioOutputDevice"),
                settings.value("sampleRate"),  # TODO: validation
                voiceChangerSettings.serverReadChunkSize * 128,
                self.changeVoice,
            )  # TODO: pass settings
            self.audio.voiceChangerFilter.passThrough = passThrough
        else:
            self.audio = None

    def changeVoice(
        self, receivedData: AudioInOutFloat
    ) -> tuple[AudioInOutFloat, float, list[int], tuple | None]:
        if self.passThrough:
            vol = float(np.sqrt(np.square(receivedData).mean(dtype=np.float32)))
            return receivedData, vol, [0, 0, 0], None

        try:
            with self.device_manager.lock:
                audio, vol, perf = self.vc.on_request(receivedData)
            return audio, vol, perf, None
        except VoiceChangerIsNotSelectedException as e:
            logger.exception(e)
            return (
                np.zeros(1, dtype=np.float32),
                0,
                [0, 0, 0],
                ("VoiceChangerIsNotSelectedException", format_exc()),
            )
        except PipelineNotInitializedException as e:
            logger.exception(e)
            return (
                np.zeros(1, dtype=np.float32),
                0,
                [0, 0, 0],
                ("PipelineNotInitializedException", format_exc()),
            )
        except Exception as e:
            logger.exception(e)
            return (
                np.zeros(1, dtype=np.float32),
                0,
                [0, 0, 0],
                ("Exception", format_exc()),
            )

    def importModel(self, params: LoadModelParams):
        slotDir = os.path.join(
            self.modelDir,
            str(params.slot),
        )

        iconFile = ""
        if os.path.isdir(slotDir):
            # Replacing existing model, delete everything except for the icon.
            slotInfo = self.modelSlotManager.get_slot_info(params.slot)
            iconFile = slotInfo.iconFile
            for entry in os.listdir(slotDir):
                if entry != iconFile:
                    filePath = os.path.join(slotDir, entry)
                    if os.path.isdir(filePath):
                        shutil.rmtree(filePath)
                    else:
                        os.remove(filePath)

        for file in params.files:
            logger.info(f"FILE: {file}")
            srcPath = os.path.join(file.dir, file.name)
            dstDir = os.path.join(
                self.modelDir,
                str(params.slot),
                file.dir,
            )
            dstPath = os.path.join(dstDir, os.path.basename(file.name))
            os.makedirs(dstDir, exist_ok=True)
            logger.info(f"Copying {srcPath} -> {dstPath}")
            shutil.copy(srcPath, dstPath)
            file.name = os.path.basename(dstPath)

        if params.voiceChangerType == "RVC":
            slotInfo = RVCModelSlotGenerator.load_model(params, self.modelDir)
            self.modelSlotManager.save_model_slot(params.slot, slotInfo)

        # Restore icon.
        slotInfo = self.modelSlotManager.get_slot_info(params.slot)
        slotInfo.iconFile = iconFile
        self.modelSlotManager.save_model_slot(params.slot, slotInfo)

        self.modelUpdated.emit(params.slot)

    def setModelIcon(self, slot: int, iconFile: str):
        iconFileBaseName = os.path.basename(iconFile)
        storePath = os.path.join(self.modelDir, str(slot), iconFileBaseName)
        try:
            shutil.copy(iconFile, storePath)
        except shutil.SameFileError:
            pass
        slotInfo = self.modelSlotManager.get_slot_info(slot)
        if slotInfo.iconFile != "" and slotInfo.iconFile != iconFileBaseName:
            os.remove(os.path.join(self.modelDir, str(slot), slotInfo.iconFile))
        slotInfo.iconFile = iconFileBaseName
        self.modelSlotManager.save_model_slot(slot, slotInfo)
        self.modelUpdated.emit(slot)


def main():
    app = QApplication(sys.argv)
    app.setDesktopFileName("AVoc")
    app.setOrganizationName("AVocOrg")
    app.setApplicationName("AVoc")

    iconFilePath = os.path.join(os.path.dirname(__file__), "AVoc.svg")

    icon = QIcon()
    icon.addFile(iconFilePath)

    app.setWindowIcon(icon)

    clParser = QCommandLineParser()
    clParser.addHelpOption()
    clParser.addVersionOption()

    noModelLoadOption = QCommandLineOption(
        ["no-model-load"], "Don't load a voice model."
    )
    clParser.addOption(noModelLoadOption)

    clParser.process(app)

    # Let Ctrl+C in terminal close the application.
    signal.signal(signal.SIGINT, lambda *args: QApplication.quit())
    timer = QTimer()
    timer.start(250)
    timer.timeout.connect(lambda: None)  # Let the interpreter run each 250 ms.

    window = MainWindow()
    window.setWindowTitle("AVoc")

    splash = QSplashScreen(QPixmap(iconFilePath))
    splash.show()  # Order is important.
    window.show()  # Order is important. And calling window.show() is important.
    window.hide()
    app.processEvents()

    # Set the path where the voice models are stored and pretrained weights are loaded.
    appLocalDataLocation = QStandardPaths.writableLocation(
        QStandardPaths.StandardLocation.AppLocalDataLocation
    )
    if appLocalDataLocation == "":
        raise FailedToSetModelDirException

    # Check or download models that used internally by the algorithm.
    pretrainDir = os.path.join(appLocalDataLocation, PRETRAIN_DIR_NAME)
    asyncio.run(downloadWeight(pretrainDir))

    # Lay out the window.
    window.initialize(os.path.join(appLocalDataLocation, MODEL_DIR_NAME))

    # Create the voice changer and connect it to the controls.
    window.vcm = VoiceChangerManager(window.windowAreaWidget.modelDir, pretrainDir)
    window.windowAreaWidget.startButton.toggled.connect(
        lambda checked: window.vcm.setRunning(
            checked, window.windowAreaWidget.passThroughButton.isChecked()
        )
    )

    def setPassThrough(passThrough: bool):
        if window.vcm.audio is not None:
            window.vcm.audio.voiceChangerFilter.passThrough = passThrough

    window.windowAreaWidget.passThroughButton.toggled.connect(setPassThrough)

    modelSettingsGroupBox = window.windowAreaWidget.modelSettingsGroupBox

    def onModelSettingsChanged():
        window.vcm.setModelSettings(
            pitch=modelSettingsGroupBox.pitchSpinBox.value(),
            formantShift=modelSettingsGroupBox.formantShiftDoubleSpinBox.value(),
            index=modelSettingsGroupBox.indexDoubleSpinBox.value(),
        )

    modelSettingsGroupBox.changed.connect(onModelSettingsChanged)

    interfaceSettings = QSettings()
    interfaceSettings.beginGroup("Interface")

    def onVoiceCardChanged() -> None:
        modelSettingsGroupBox.changed.disconnect(onModelSettingsChanged)
        window.vcm.initialize()
        modelSettingsGroupBox.changed.connect(onModelSettingsChanged)
        if bool(interfaceSettings.value("showNotifications", True)):
            voiceCardWidget: QLabel = window.windowAreaWidget.voiceCards.itemWidget(
                window.windowAreaWidget.voiceCards.currentItem()
            )
            pixmap = voiceCardWidget.pixmap()
            window.showTrayMessage(
                window.windowTitle(),
                f"Switched to {voiceCardWidget.toolTip()}",
                pixmap,
            )

    def onModelSettingsLoaded(pitch: int, formantShift: float, index: float):
        modelSettingsGroupBox.pitchSpinBox.setValue(pitch)
        modelSettingsGroupBox.formantShiftDoubleSpinBox.setValue(formantShift)
        modelSettingsGroupBox.indexDoubleSpinBox.setValue(index)

    window.vcm.modelSettingsLoaded.connect(onModelSettingsLoaded)

    window.windowAreaWidget.voiceCards.currentRowChanged.connect(onVoiceCardChanged)
    window.windowAreaWidget.voiceCards.droppedModelFiles.connect(
        lambda loadModelParams: window.vcm.importModel(loadModelParams)
    )
    window.windowAreaWidget.voiceCards.droppedIconFile.connect(
        lambda slot, iconFile: window.vcm.setModelIcon(slot, iconFile)
    )
    window.vcm.modelUpdated.connect(
        lambda slot: window.windowAreaWidget.voiceCards.onVoiceCardUpdated(slot),
    )

    audioSettingsGroupBox = window.windowAreaWidget.audioSettingsGroupBox
    audioSettingsGroupBox.sampleRateComboBox.currentIndexChanged.connect(
        lambda: window.vcm.initialize()
    )  # It isn't running when changing sample rate.

    # Load the current voice model if any.
    if not clParser.isSet(noModelLoadOption):
        window.vcm.initialize()

    # Show the window
    window.resize(1980, 1080)  # TODO: store interface dimensions
    window.show()
    splash.finish(window)

    sys.exit(app.exec())



from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import (
    QDoubleSpinBox,
    QGridLayout,
    QGroupBox,
    QLabel,
    QSlider,
    QSpinBox,
    QWidget,
)


class ModelSettingsGroupBox(QGroupBox):
    changed = Signal()

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)

        self.setTitle("Settings for the Active Voice Model")

        modelSettingsLayout = QGridLayout()
        row = 0
        pitchLabel = QLabel("Pitch")
        pitchLabel.setToolTip(
            "Pitch shift alters the tone of voice. If your voice is low-pitched (f.e., male voice) and model is high-pitched (f.e., a female voice model), increase this setting (i.e., set to 12). Otherwise, decrease this setting (i.e., set to -12)."  # noqa: E501
        )
        pitchSlider = QSlider(Qt.Orientation.Horizontal)
        self.pitchSpinBox = QSpinBox(minimum=-50, maximum=50)
        pitchSlider.setMinimum(self.pitchSpinBox.minimum())
        pitchSlider.setMaximum(self.pitchSpinBox.maximum())
        pitchSlider.valueChanged.connect(lambda v: self.pitchSpinBox.setValue(v))
        self.pitchSpinBox.valueChanged.connect(lambda v: pitchSlider.setValue(v))
        self.pitchSpinBox.valueChanged.connect(lambda: self.changed.emit())
        modelSettingsLayout.addWidget(pitchLabel, row, 0)
        modelSettingsLayout.addWidget(pitchSlider, row, 1)
        modelSettingsLayout.addWidget(self.pitchSpinBox, row, 2)
        row += 1
        formantShiftLabel = QLabel("Formant Shift")
        formantShiftLabel.setToolTip(
            "Formant shift alters harmonic frequencies and changes the voice timbre without affecting the pitch."  # noqa: E501
        )
        formantShiftSlider = QSlider(Qt.Orientation.Horizontal)
        self.formantShiftDoubleSpinBox = QDoubleSpinBox(
            minimum=-5, maximum=5, singleStep=0.1, decimals=1
        )
        formantShiftSlider.setMinimum(
            int(self.formantShiftDoubleSpinBox.minimum() * 10)
        )
        formantShiftSlider.setMaximum(
            int(self.formantShiftDoubleSpinBox.maximum() * 10)
        )
        formantShiftSlider.valueChanged.connect(
            lambda v: self.formantShiftDoubleSpinBox.setValue(v / 10.0)
        )
        self.formantShiftDoubleSpinBox.valueChanged.connect(
            lambda v: formantShiftSlider.setValue(v * 10)
        )
        self.formantShiftDoubleSpinBox.valueChanged.connect(lambda: self.changed.emit())
        modelSettingsLayout.addWidget(formantShiftLabel, row, 0)
        modelSettingsLayout.addWidget(formantShiftSlider, row, 1)
        modelSettingsLayout.addWidget(self.formantShiftDoubleSpinBox, row, 2)
        row += 1
        indexLabel = QLabel("Index")
        indexLabel.setToolTip(
            "Index embeds accent of the model's voice into your voice. Disabled when set to 0. Note that this setting increases CPU usage."  # noqa: E501
        )
        indexSlider = QSlider(Qt.Orientation.Horizontal)
        self.indexDoubleSpinBox = QDoubleSpinBox(
            minimum=0, maximum=1, singleStep=0.1, decimals=1
        )
        indexSlider.setMinimum(int(self.indexDoubleSpinBox.minimum() * 10))
        indexSlider.setMaximum(int(self.indexDoubleSpinBox.maximum() * 10))
        indexSlider.valueChanged.connect(
            lambda v: self.indexDoubleSpinBox.setValue(v / 10.0)
        )
        self.indexDoubleSpinBox.valueChanged.connect(
            lambda v: indexSlider.setValue(v * 10)
        )
        self.indexDoubleSpinBox.valueChanged.connect(lambda: self.changed.emit())
        modelSettingsLayout.addWidget(indexLabel, row, 0)
        modelSettingsLayout.addWidget(indexSlider, row, 1)
        modelSettingsLayout.addWidget(self.indexDoubleSpinBox, row, 2)

        self.setLayout(modelSettingsLayout)



from PySide6.QtCore import QByteArray
from PySide6.QtMultimedia import QAudioFormat, QMediaDevices


def getAudioDevicesForSampleRate(
    sampleRate: int, isInput: bool
) -> dict[QByteArray, (str, bool)]:
    """Return compatible devices: id, readable name, a flag if it's the default device."""

    if isInput:
        audioDevices = QMediaDevices.audioInputs()
    else:
        audioDevices = QMediaDevices.audioOutputs()

    format = QAudioFormat()
    format.setSampleRate(sampleRate)
    format.setChannelCount(1)
    format.setSampleFormat(QAudioFormat.SampleFormat.Float)

    return {
        d.id(): (d.description(), d.isDefault())
        for d in audioDevices
        if d.isFormatSupported(format)
    }



from PySide6.QtCore import QByteArray, QSettings
from PySide6.QtWidgets import QComboBox, QGridLayout, QGroupBox, QLabel, QWidget

from .audiodevices import getAudioDevicesForSampleRate

DEFAULT_SAMPLE_RATE = 48000
SAMPLE_RATES = (
    32000,
    44100,
    48000,
    96000,
    128000,
)  # TODO: 44100 leads to crashes because the algo resamples to 16k badly


class AudioSettingsGroupBox(QGroupBox):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)

        self.setTitle("Audio Settings")

        audioSettingsLayout = QGridLayout()
        audioSettingsLayout.setColumnStretch(1, 1)  # Stretch the comboboxes.
        row = 0
        self.sampleRateComboBox = QComboBox()
        audioSettingsLayout.addWidget(QLabel("Sample Rate"), row, 0)
        audioSettingsLayout.addWidget(self.sampleRateComboBox, row, 1)
        row += 1
        self.audioInputComboBox = AudioDevicesComboBox(isInput=True)
        self.audioInputComboBox.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon
        )
        audioSettingsLayout.addWidget(QLabel("Audio Input"), row, 0)
        audioSettingsLayout.addWidget(self.audioInputComboBox, row, 1)
        row += 1
        self.audioOutputComboBox = AudioDevicesComboBox(isInput=False)
        self.audioOutputComboBox.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon
        )
        audioSettingsLayout.addWidget(QLabel("Audio Output"), row, 0)
        audioSettingsLayout.addWidget(self.audioOutputComboBox, row, 1)

        self.setLayout(audioSettingsLayout)

        settings = QSettings()
        settings.beginGroup("AudioSettings")

        # Restore the sample rate from saved settings.
        for sampleRate in SAMPLE_RATES:
            self.sampleRateComboBox.addItem(str(sampleRate))

        savedSampleRate = settings.value("sampleRate", DEFAULT_SAMPLE_RATE)
        index = (
            self.sampleRateComboBox.findText(str(savedSampleRate))
            if savedSampleRate is not None
            else -1
        )
        if index >= 0:
            self.sampleRateComboBox.setCurrentIndex(index)
        else:
            self.sampleRateComboBox.setCurrentIndex(
                SAMPLE_RATES.index(DEFAULT_SAMPLE_RATE)
            )

        def onCurrentIndexChanged(index: int):
            self.audioInputComboBox.refreshDeviceOptions(SAMPLE_RATES[index])
            self.audioOutputComboBox.refreshDeviceOptions(SAMPLE_RATES[index])

        self.sampleRateComboBox.currentIndexChanged.connect(onCurrentIndexChanged)

        # Load the device lists.
        onCurrentIndexChanged(self.sampleRateComboBox.currentIndex())

        # Restore the input device from saved settings.
        self.audioInputComboBox.restoreFromSavedSetting(
            settings.value("audioInputDevice")
        )
        self.audioOutputComboBox.restoreFromSavedSetting(
            settings.value("audioOutputDevice")
        )

        # Set up the saving of the settings.
        self.sampleRateComboBox.currentIndexChanged.connect(
            lambda: settings.setValue(
                "sampleRate", int(self.sampleRateComboBox.currentText())
            ),
        )
        self.audioInputComboBox.currentIndexChanged.connect(
            lambda: settings.setValue(
                "audioInputDevice", self.audioInputComboBox.currentData()
            ),
        )
        self.audioOutputComboBox.currentIndexChanged.connect(
            lambda: settings.setValue(
                "audioOutputDevice", self.audioOutputComboBox.currentData()
            ),
        )

        # Sync settings in case they weren't initialized.
        settings.setValue("sampleRate", int(self.sampleRateComboBox.currentText()))
        settings.setValue("audioInputDevice", self.audioInputComboBox.currentData())
        settings.setValue("audioOutputDevice", self.audioOutputComboBox.currentData())


class AudioDevicesComboBox(QComboBox):
    def __init__(self, isInput: bool, parent: QWidget | None = None):
        super().__init__(parent)
        self.isInput = isInput
        self.defaultIndex = 0

    def refreshDeviceOptions(self, sampleRate: int):
        audioDevices = getAudioDevicesForSampleRate(sampleRate, self.isInput)
        if sorted(audioDevices.keys()) != sorted(
            [self.itemData(i) for i in range(self.count())]
        ):
            self.clear()
            index = 0
            for id, (description, isDefault) in audioDevices.items():
                self.addItem(description, id)
                if isDefault:
                    self.defaultIndex = index
                index += 1

    def restoreFromSavedSetting(self, id: QByteArray | None):
        index = self.findData(id) if id is not None else -1
        self.setCurrentIndex(index if index >= 0 else self.defaultIndex)



import numpy as np
from PySide6.QtCore import QByteArray, QIODevice
from PySide6.QtMultimedia import (
    QAudioDevice,
    QAudioFormat,
    QAudioSink,
    QAudioSource,
    QMediaDevices,
)

from .exceptions import AudioDeviceDisappearedException


class AudioFilter(QIODevice):
    def __init__(
        self, inputDevice: QIODevice, blockSamplesCount: int, changeVoice, parent=None
    ):
        super().__init__(parent)

        self.passThrough = False
        self.inputDevice = inputDevice
        self.inputDevice.readyRead.connect(self.onReadyRead)
        self.changeVoice = changeVoice
        self.audioInBuff = np.empty(0, dtype=np.float32)
        self.blockSamplesCount = blockSamplesCount

    def readData(self, maxlen: int) -> object:
        if maxlen == 0:
            return b""

        self.audioInBuff = np.append(
            self.audioInBuff,
            np.frombuffer(self.inputDevice.read(maxlen), dtype=np.float32),
        )

        # Use the device buffer limit as an opportunity to catch up.
        self.audioInBuff = self.audioInBuff[-maxlen // 4 :]

        result = np.empty(0, dtype=np.float32)

        if self.passThrough:
            result = self.audioInBuff
            self.audioInBuff = np.empty(0, dtype=np.float32)
            return result.tobytes()

        blockCount = len(self.audioInBuff) // self.blockSamplesCount

        for blockIndex in range(0, blockCount):
            bs = blockIndex * self.blockSamplesCount
            be = bs + self.blockSamplesCount
            out_wav, _, _, _ = self.changeVoice(self.audioInBuff[bs:be])
            result = np.append(result, out_wav)

        return result.tobytes()

    def isSequential(self) -> bool:
        return self.inputDevice.isSequential()

    def onReadyRead(self):
        if self.bytesAvailable() != 0:
            self.readyRead.emit()

    def bytesAvailable(self) -> int:
        srcBytesCount = len(self.audioInBuff) * 4 + self.inputDevice.bytesAvailable()
        available = srcBytesCount - srcBytesCount % (self.blockSamplesCount * 4)
        return available


def getAudioDeviceById(deviceId: QByteArray, isInput: bool) -> QAudioDevice:
    devices = QMediaDevices.audioInputs() if isInput else QMediaDevices.audioOutputs()

    for dev in devices:
        if dev.id() == deviceId:
            return dev

    raise AudioDeviceDisappearedException


class Audio:
    def __init__(
        self,
        audioInputDeviceId: QByteArray,
        audioOutputDeviceId: QByteArray,
        sampleRate: int,
        blockSamplesCount: int,
        changeVoice,
    ):
        audioInputDevice = getAudioDeviceById(
            audioInputDeviceId, isInput=True
        )  # TODO: exception
        audioInputFormat = audioInputDevice.preferredFormat()
        audioInputFormat.setSampleRate(sampleRate)
        audioInputFormat.setSampleFormat(QAudioFormat.SampleFormat.Float)
        self.audioSource = QAudioSource(
            audioInputDevice,
            audioInputFormat,
        )  # TODO: check opening

        audioOutputDevice = getAudioDeviceById(
            audioOutputDeviceId, isInput=False
        )  # TODO: exception
        self.audioSink = QAudioSink(
            audioOutputDevice,
            audioInputFormat,
        )  # TODO: check opening

        # Start the IO.
        self.voiceChangerFilter = AudioFilter(
            self.audioSource.start(),
            blockSamplesCount,
            changeVoice,
        )  # TODO: check audioSource.error()
        self.voiceChangerFilter.open(
            QIODevice.OpenModeFlag.ReadOnly
        )  # TODO: check opening

        # Do the loopback.
        self.audioSink.start(self.voiceChangerFilter)  # TODO: check audioSink.error()

        # TODO: connect slots to audioSink/audioSource errors to catch device changes.



from PySide6.QtCore import QSettings, Signal
from PySide6.QtWidgets import QCheckBox, QPushButton, QVBoxLayout, QWidget


class CustomizeUiWidget(QWidget):
    back = Signal()

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)

        interfaceSettings = QSettings()
        interfaceSettings.beginGroup("Interface")

        layout = QVBoxLayout()

        self.showNotificationsCheckBox = QCheckBox("Show Notifications")
        self.showNotificationsCheckBox.setChecked(
            bool(interfaceSettings.value("showNotifications", True))
        )
        self.showNotificationsCheckBox.toggled.connect(
            lambda checked: interfaceSettings.setValue("showNotifications", checked)
        )

        layout.addWidget(self.showNotificationsCheckBox)

        self.backButton = QPushButton("Back")
        self.backButton.clicked.connect(lambda: self.back.emit())
        layout.addWidget(self.backButton)

        layout.addStretch()

        self.setLayout(layout)



import json
import os
import re
import shutil
from typing import Iterable, List

from PySide6.QtCore import QModelIndex, QSettings, QSize, Qt, Signal
from PySide6.QtGui import (
    QDragEnterEvent,
    QDragMoveEvent,
    QDropEvent,
    QFontMetrics,
    QImageReader,
    QPalette,
    QPixmap,
)
from PySide6.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QLabel,
    QListView,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from voiceconversion.utils.LoadModelParams import LoadModelParamFile, LoadModelParams

from .audiosettings import AudioSettingsGroupBox
from .exceptions import FailedToMoveVoiceCardException
from .modelsettings import ModelSettingsGroupBox

VOICE_CARD_SIZE = QSize(188, 262)
VOICE_CARD_MARGIN = 8

UNKNOWN_MODEL_NAME = "Unknown Model"
DROP_MODEL_FILES = "Drop model files here<br><b>*.pth</b> and <b>*.index</b><br><br>"
DROP_ICON_FILE = "Drop icon file here<br><b>*.png</b>, <b>*.jpeg</b>, <b>*.gif</b>..."
START_TXT = "Start"
RUNNING_TXT = "Running..."
PASS_THROUGH_TXT = "Pass Through"


def voiceCardForSlot(modelDir: str, row: int) -> QWidget:
    folder = str(row)
    paramsFilePath = os.path.join(modelDir, folder, "params.json")
    if os.path.exists(paramsFilePath):
        name = UNKNOWN_MODEL_NAME
        with open(paramsFilePath) as f:
            params = json.load(f)
            iconFileName = params.get("iconFile", "")
            name = params.get("name", name)
            if iconFileName:
                pixmap = QPixmap(os.path.join(modelDir, folder, iconFileName))
                basename = os.path.splitext(os.path.basename(iconFileName))[0]
                label = QLabel()
                label.setPixmap(cropCenterScalePixmap(pixmap, VOICE_CARD_SIZE))
                label.setToolTip(basename)
                return label
    return VoiceCardPlaceholderWidget(
        VOICE_CARD_SIZE, f"{name}<br><br>{DROP_ICON_FILE}"
    )


class WindowAreaWidget(QWidget):
    def __init__(self, modelDir: str, parent: QWidget | None = None):
        super().__init__(parent)

        self.modelDir = modelDir

        settings = QSettings()
        settings.beginGroup("Interface")

        layout = QVBoxLayout()

        self.voiceCards = VoiceCardsContainer(modelDir)

        layout.addWidget(self.voiceCards, stretch=2)

        controlsLayout = QHBoxLayout()

        self.audioSettingsGroupBox = AudioSettingsGroupBox()
        controlsLayout.addWidget(self.audioSettingsGroupBox, stretch=3)

        self.modelSettingsGroupBox = ModelSettingsGroupBox()
        controlsLayout.addWidget(self.modelSettingsGroupBox, stretch=1)

        buttonsLayout = QVBoxLayout()

        self.startButton = QPushButton(START_TXT)
        fm = QFontMetrics(self.startButton.font())
        maxButtonWidth = int(
            max(
                fm.horizontalAdvance(t)
                for t in [START_TXT, RUNNING_TXT, PASS_THROUGH_TXT]
            )
            * 1.618
        )
        # Make the Start button size fixed.
        self.startButton.setMinimumWidth(maxButtonWidth)
        # Make the Start button toggle and change text when clicked.
        self.startButton.setCheckable(True)
        self.startButton.toggled.connect(
            lambda checked: self.startButton.setText(
                RUNNING_TXT if checked else START_TXT
            )
        )
        # Unfortunately can't drag the cards while the voice conversion is running
        # because it will select them and load.
        self.startButton.toggled.connect(
            lambda checked: self.voiceCards.setDragDropMode(
                QAbstractItemView.DragDropMode.DropOnly
                if checked
                else QAbstractItemView.DragDropMode.InternalMove
            )
        )
        # Can't change audio settings while running.
        self.startButton.toggled.connect(
            lambda checked: self.audioSettingsGroupBox.setEnabled(not checked)
        )
        buttonsLayout.addWidget(self.startButton)

        self.passThroughButton = QPushButton(PASS_THROUGH_TXT)
        self.passThroughButton.setMinimumWidth(maxButtonWidth)
        self.passThroughButton.setCheckable(True)
        passThrough = settings.value("passThrough", False, type=bool)
        assert type(passThrough) is bool
        self.passThroughButton.setChecked(passThrough)
        self.passThroughButton.toggled.connect(
            lambda checked: settings.setValue("passThrough", checked)
        )
        buttonsLayout.addWidget(self.passThroughButton)

        controlsLayout.addLayout(buttonsLayout)

        layout.addLayout(controlsLayout, stretch=1)

        self.setLayout(layout)

        modelDirToModelIcon: dict[str, QWidget] = {}

        os.makedirs(modelDir, exist_ok=True)

        for folder in os.listdir(modelDir):
            if folder.isdigit() and os.path.isdir(os.path.join(modelDir, folder)):
                modelDirToModelIcon[folder] = voiceCardForSlot(modelDir, int(folder))

        for folder in sortedNumerically(modelDirToModelIcon):
            self.voiceCards.addWidget(modelDirToModelIcon[folder])

        self.voiceCards.addWidget(
            VoiceCardPlaceholderWidget(
                VOICE_CARD_SIZE, DROP_MODEL_FILES + DROP_ICON_FILE
            ),
            selectable=False,
        )

        self.voiceCards.model().rowsMoved.connect(self.rearrangeVoiceModelDirs)

        self.voiceCards.setCurrentRow(int(settings.value("currentVoiceCardIndex", 0)))
        self.voiceCards.currentRowChanged.connect(
            lambda row: settings.setValue("currentVoiceCardIndex", row)
        )

        # Disable the start button if there are no voice cards.
        self.voiceCards.currentRowChanged.connect(
            lambda row: self.startButton.setEnabled(row >= 0)
        )
        self.startButton.setEnabled(self.voiceCards.currentRow() >= 0)

    def rearrangeVoiceModelDirs(
        self,
        sourceParent: QModelIndex,
        sourceStart: int,
        sourceEnd: int,
        destinationParent: QModelIndex,
        destinationRow: int,
    ):
        if sourceStart != sourceEnd:
            raise FailedToMoveVoiceCardException

        dirs = sorted(
            [d for d in os.listdir(self.modelDir) if d.isdigit()], key=lambda x: int(x)
        )
        total = len(dirs)

        if not (0 <= sourceStart < total) or not (0 <= destinationRow <= total):
            raise FailedToMoveVoiceCardException("Invalid indices")

        if destinationRow > sourceStart:
            destinationRow -= 1

        # Create a temp name for the moving directory to avoid name conflicts
        src_path = os.path.join(self.modelDir, str(sourceStart))
        tmp_path = os.path.join(self.modelDir, "_tmp_move")
        shutil.move(src_path, tmp_path)

        # Renumber other directories depending on move direction
        if sourceStart < destinationRow:
            # Shift everything between (sourceStart+1 ... destinationRow) up by one
            for i in range(sourceStart + 1, destinationRow + 1):
                os.rename(
                    os.path.join(self.modelDir, str(i)),
                    os.path.join(self.modelDir, str(i - 1)),
                )
        else:
            # Shift everything between (destinationRow ... sourceStart-1) down by one
            for i in range(sourceStart - 1, destinationRow - 1, -1):
                os.rename(
                    os.path.join(self.modelDir, str(i)),
                    os.path.join(self.modelDir, str(i + 1)),
                )

        # Move the temp folder into its new numbered slot
        shutil.move(tmp_path, os.path.join(self.modelDir, str(destinationRow)))


class FlowContainer(QListWidget):
    def __init__(self):
        super().__init__()

        # Allow dragging the cards around
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)

        # make it look like a normal scroll area
        self.viewport().setBackgroundRole(QPalette.Window)
        # display items from left to right, instead of top to bottom
        self.setFlow(QListView.Flow.LeftToRight)
        # wrap items that don't fit the width of the viewport
        # similar to self.setViewMode(self.IconMode)
        self.setWrapping(True)
        # always re-layout items when the view is resized
        self.setResizeMode(QListView.ResizeMode.Adjust)

        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setVerticalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        # Add margins for the items to make the selection frame around a card visible.
        self.setStyleSheet(
            f"""
            QListWidget::item {{
                margin:{VOICE_CARD_MARGIN}px;
            }}
            """
        )

    def addWidget(self, widget: QWidget, selectable: bool = True):
        item = QListWidgetItem()
        if not selectable:
            item.setFlags(
                item.flags()
                & ~(Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled)
            )
        self.addItem(item)
        item.setSizeHint(widget.sizeHint())
        self.setItemWidget(item, widget)

    def insertWidget(self, row: int, widget: QWidget):
        item = QListWidgetItem()
        self.insertItem(row, item)
        item.setSizeHint(widget.sizeHint())
        self.setItemWidget(item, widget)


class FlowContainerWithFixedLast(FlowContainer):
    def canDropBeforLast(self, event: QDropEvent):
        """Forbid going past the last item which is the voice card placeholder."""
        row = self.indexAt(event.pos()).row()
        if row == self.count() - 1:
            itemRect = self.visualRect(self.model().index(self.count() - 1, 0))
            return event.pos().x() < itemRect.center().x()
        return row > 0

    def dragMoveEvent(self, event: QDragMoveEvent):
        if self.canDropBeforLast(event):
            super().dragMoveEvent(event)

    def dropEvent(self, event: QDropEvent):
        # InternalMove drops to rearrange cards.
        if self.canDropBeforLast(event):
            super().dropEvent(event)
        else:
            # Hack to clear a failed drop indicator
            self.setDropIndicatorShown(False)
            self.viewport().update()
            self.setDropIndicatorShown(True)


class VoiceCardsContainer(FlowContainerWithFixedLast):

    droppedModelFiles = Signal(LoadModelParams)
    droppedIconFile = Signal(int, str)

    def __init__(self, modelDir: str):
        super().__init__()
        self.modelDir = modelDir

    def dragEnterEvent(self, event: QDragEnterEvent):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            super().dragEnterEvent(event)

    def dropEvent(self, event: QDropEvent):
        if not event.mimeData().hasUrls():
            # InternalMove drops to rearrange cards.
            return super().dropEvent(event)

        # External drop to import models and icons.

        files = [url for url in event.mimeData().urls() if url.isLocalFile()]
        pthFiles = [file for file in files if file.toString().endswith(".pth")]
        indexFiles = [file for file in files if file.toString().endswith(".index")]
        iconFiles = [
            file
            for file in files
            if not file.toString().endswith(".pth")
            and not file.toString().endswith(".index")
        ]
        if len(pthFiles) != len(indexFiles):
            QMessageBox.critical(
                self,
                "Error Importing a Voice Model",
                "Both files expected: *.pth and *.index",
            )
            return
        if len(indexFiles) == 1:
            indexFile = indexFiles[0].toLocalFile()
            if os.path.basename(indexFile).startswith("trained"):
                QMessageBox.critical(
                    self,
                    "Error Importing a Voice Model",
                    f"Use the 'added' index, not the 'trained'.\n\nFile:\n{indexFile}",
                )
                return
        if len(iconFiles) > 1:
            QMessageBox.critical(
                self,
                "Error Importing an Icon",
                "Only one icon file expected.",
            )
        if len(iconFiles) == 1:
            iconFile = iconFiles[0].toLocalFile()
            supportedImageFormats = QImageReader.supportedImageFormats()
            iconFileExt = os.path.splitext(iconFile)[1][1:].lower()
            if iconFileExt.encode("utf-8") not in supportedImageFormats:
                QMessageBox.critical(
                    self,
                    "Error Importing an Icon",
                    f"Failed to import an icon for a voice card.\n\nFile:\n{iconFile}",
                )
                return

        row = self.indexAt(event.pos()).row()
        importingNew = row < 0 or row >= self.count()
        slot = self.count() - 1 if importingNew else row

        if len(indexFiles) == 1:
            self.droppedModelFiles.emit(
                LoadModelParams(
                    voiceChangerType="RVC",
                    slot=slot,
                    isSampleMode=False,
                    sampleId="",
                    files=[
                        LoadModelParamFile(
                            name=pthFiles[0].toLocalFile(), kind="rvcModel", dir=""
                        ),
                        LoadModelParamFile(
                            name=indexFiles[0].toLocalFile(), kind="rvcIndex", dir=""
                        ),
                    ],
                    params={},
                )
            )

        if len(iconFiles) == 1 and slot < self.count() - 1:
            self.droppedIconFile.emit(slot, iconFiles[0].toLocalFile())

    def onVoiceCardUpdated(self, row: int):
        if row >= self.count() - 1:
            self.insertWidget(row, voiceCardForSlot(self.modelDir, row))
        else:
            self.setItemWidget(self.item(row), voiceCardForSlot(self.modelDir, row))


class VoiceCardPlaceholderWidget(QWidget):
    def __init__(self, cardSize: QSize, text: str, parent: QWidget | None = None):
        super().__init__(parent)

        self.cardSize = cardSize
        self.setStyleSheet("border: 2px solid;")

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        dropHere = QLabel(text)
        dropHere.setTextFormat(Qt.TextFormat.RichText)
        dropHere.setWordWrap(True)
        dropHere.setAlignment(Qt.AlignmentFlag.AlignCenter)

        layout.addWidget(dropHere)
        self.setLayout(layout)

    def sizeHint(self):
        return self.cardSize


def cropCenterScalePixmap(pixmap: QPixmap, targetSize: QSize) -> QPixmap:
    # Original size
    ow = pixmap.width()
    oh = pixmap.height()

    # Maintain target ratio
    target_ratio = targetSize.width() / targetSize.height()
    orig_ratio = ow / oh

    if orig_ratio > target_ratio:
        # Original is too wide → crop horizontally
        cropW = int(oh * target_ratio)
        cropH = oh
        x = (ow - cropW) // 2  # center horizontally
        y = 0  # from top
    else:
        # Original is too tall → crop vertically
        cropW = ow
        cropH = int(ow / target_ratio)
        x = 0
        y = 0  # from top (not centered vertically)

    cropped = pixmap.copy(x, y, cropW, cropH)

    return cropped.scaled(targetSize, mode=Qt.TransformationMode.SmoothTransformation)


def sortedNumerically(input: Iterable[str]) -> List[str]:
    def repl(num):
        return f"{int(num[0]):010d}"

    return sorted(input, key=lambda i: re.sub(r"(\d+)", repl, i))



class VoiceChangerIsNotSelectedException(Exception):
    def __str__(self):
        return repr("Voice Changer is not selected.")


class PipelineNotInitializedException(Exception):
    def __str__(self):
        return repr("Pipeline is not initialized.")


class AudioDeviceDisappearedException(Exception):
    def __str__(self):
        return repr("Audio device disappeared.")


class FailedToMoveVoiceCardException(Exception):
    def __str__(self):
        return repr("Failed to move voice card.")


class FailedToSetModelDirException(Exception):
    def __str__(self):
        return repr("Failed set the directory for storing the voice models.")
